#define USECHRONO
#include <cassert>
#include <climits>
#include <cstdlib>
#include <stack>
#include <Eigen/Dense>
#include "eval.hpp"

using namespace aed;
using namespace std;
using namespace Eigen;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
typedef int (*mapfun_t)(int);
void separate(set<int> &S,mapfun_t f,map<int,set<int> > &f2x) {
  // COMPLETAR...
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
// Clase virtual pura que describe grafos
class graph_t {
public:
  virtual int size()=0;
  virtual int dist(int i,int j)=0;
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
bool next_path(vector<int> &path) {
  // Dado un camino en el grafo lo avanza al siguiente
  // camino. Aplicando sucesivamente `next_path' se pueden
  // recorrer todos los caminos
  int N = path.size();
  for (int j=N-1; j>=0; j--) {
    if (path[0]==j) {
      for (int k=0; k<j; k++) path[k] = path[k+1];
      path[j] = j;
    } else {
      int k;
      for (k=0; k<N-1; k++) 
	if (path[k]==j) break;
      path[k] = path[k-1];
      path[k-1] = j;
      return 1;
    }
  }
  return 0;
}

// Función que encuentra el mínimo recorrido en un grafo
// evaluando todos los posibles caminos con el uso de next_path
void min_path(graph_t &G,vector<int> &pathmin) {
  // COMPLETAR...
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
// Ejemplo de clase derivada. Es un grafo con N vértices
// donde el vertice j está unido al siguiente j+1
// (cíclicamente, es decir con (j+1)%N)
class circle_t : public graph_t {
public:
  int N;
  int size() { return N; }
  int dist(int i,int j) {
    if (i==(j+1)%N || j==(i+1)%N) return 1;
    else return 10000;
  }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
// Función auxiliar que imprime un camino
void ppath(const vector<int> &path) {
  for (auto q:path) cout << q << " ";
  cout << endl;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
// Función ejemplo que recorre todos los caminos y
// los imprime
void print_all_paths() {
  int N = 4;
  vector<int> path(N);
  for (int j=0; j<N; j++) path[j] = j;
  int k=0;
  while (1) {
    cout << "camino " << k++ << ": ";
    ppath(path);
    if (!next_path(path)) break;
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void min_path(graph_t &G,vector<int> &pathmin) {
  // COMPLETAR...
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void check_min_path() {
  cout << "-------------------" << endl
       << "CHECK MINPATH" << endl;
  vector<int> path;
  int minlen;

  circle_t G;
  G.N=4;
  minlen = min_path(G,path);
  cout << "circle graph(4): ";
  ppath(path);
  cout << "min length: " << minlen << endl;

#if 0  
  G.N=7;
  minlen = min_path(G,path);
  cout << "circle graph(7): ";
  ppath(path);
  cout << "min length: " << minlen << endl;
  
  full_t G2;
  G2.N=4;
  minlen = min_path(G2,path);
  cout << "full graph: ";
  ppath(path);
  cout << "min length: " << minlen << endl;

  grid_t G3;
  G3.Nx=3;
  G3.Ny=3;
  minlen = min_path(G3,path);
  cout << "grid graph 3x3: ";
  ppath(path);
  cout << "min length: " << minlen << endl;

  G3.Nx=4;
  G3.Ny=2;
  minlen = min_path(G3,path);
  cout << "grid graph 4x2: ";
  ppath(path);
  cout << "min length: " << minlen << endl;
#endif
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void mkortho(MatrixXd &Q) {
  // COMPLETAR...
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void check_mkortho() {
  cout << "-------------------" << endl
       << "CHECK MKORTHO" << endl;
  for (int z=0; z<10; z++) {
    int N=3+rand()%4;
    MatrixXd Q(N,N),Q0;
    for (int j=0; j<N; j++) {
      for (int k=0; k<N; k++) 
        Q(j,k) = 0.1*drand();
      Q(j,j) += 1.0;
    }
    Q0=Q;
    // cout << Q << endl;
    mkortho(Q);
    // cout << Q << endl;
    MatrixXd DQ = Q-Q0;
    double adq = DQ.norm();
    MatrixXd Id = MatrixXd::Identity(N,N);
    MatrixXd E = Q*Q.transpose()-Id;
    double erro = E.norm();
    cout << "correction: " << adq << ", erro " << erro << endl;
  }
}  

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  // Para chequear ejercicio 1
  Eval ev;
  ev.eval<1>(separate,vrbs);

  // Para chequear ejercicio 2
  // Para probar el recorrido de los caminos
  // print_all_paths();
  check_min_path();

  // Para chequear ejercicio 3
  check_mkortho();

  return 0;
}
