// GITVERSION: fsiconf.0-10-ga5a3c317
#ifndef EVAL_HPP
#define EVAL_HPP
#include <cassert>
#include <cstdlib>
#include <cstdlib>
#include <Eigen/Dense>
#include "aedtools/evalbase.hpp"

using namespace aed;
using namespace std;
using namespace Eigen;


bool sumzero(set<int> &);
bool hassize2(set<int> &);
bool hasodd(set<int> &);
bool allpositive(set<int> &);

namespace aed {

  list<int>::iterator string_to_iterator(list<int> &L, string str) {
    if (str=="begin") return L.begin();
    if (str=="end") return L.end();
    assert(str.find("begin+")==0);
    list<int>::iterator it = L.begin();
    advance(it,atoi(str.substr(6).c_str()));
    return it;
  }
  
  string iterator_to_string(list<int> &L, list<int>::iterator it) {
    if (it==L.end()) return "end";
    int n = std::distance(L.begin(),it);
    if (n==0) return "begin";
    return string("begin+")+to_string(n);
  }
  
  list<int>::iterator my_advance(list<int> &L, int n) {
    list<int>::iterator it = L.begin();
    std::advance(it,n);
    return it;
  }
  
  typedef map<int,set<int> > map_t;
  void json2map(json &J,map_t &M) {
    for (json::iterator q = J.begin(); q != J.end(); ++q) {
      int k = (*q)["f"];
      set<int> x = (*q)["xset"];
      M[k] = x;
    }
  }

  void map2json(map_t &M,json &J) {
    auto q = M.begin();
    while (q!=M.end()) {
      json z;
      z["f"] = q->first;
      z["xset"] = q->second;
      J.push_back(z);
      q++;
    }
  }

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> EJERC2 SPLITMONO
  typedef void (*splitmono_t)(vector<int> &v,vector< vector<int> > &vmono);
  class eval1_t : public eval_base_t {
  public:
    splitmono_t F;
    eval1_t() { dumptests=0; F=NULL; ejerc=1; testfile="./sm.json"; }
    void run_case(json &data,json &outuser) {
      // Read v
      vector<int> v = data["v"];
      vector< vector<int> > vmono;
      F(v,vmono);
      for (auto z:vmono) {
        json jz = z;
        outuser.push_back(jz);
      }
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int N = 8+rnd.rand()%8;
      vector<int> v(N);
      for (int j=0; j<N; j++) v[j] = rnd.rand()%5;
      datain["v"] = v;
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> EJERC2 FILTERSETS  
  typedef bool (*pred_t)(set<int> &S);
  
  typedef void (*filtersets_t)(vector< set<int> > &vset,
                               vector< set<int> > &vsetout,
                               pred_t pred);
  class eval2_t : public eval_base_t {
  public:
    filtersets_t F;
    eval2_t() { dumptests=0; F=NULL; ejerc=2; testfile="./fs.json"; }
    void run_case(json &data,json &outuser) {
      vector< set<int> > vsetin;
      json vs = data["vsetin"];
      for (auto q:vs) {
        set<int> s = q;
        vsetin.push_back(s);
      }
      string p = data["pred"];
      
      pred_t pred;
#define CHECK(z) else if (p==#z) pred=z
      if (0) {}
      CHECK(sumzero);
      CHECK(hassize2);
      CHECK(hasodd);
      CHECK(allpositive);
      else throw string("can't find function ") + p;
      vector< set<int> > vsetout;
      F(vsetin,vsetout,pred);
      for (auto z:vsetout) {
        json jz = z;
        outuser.push_back(jz);
      }
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int k=rnd.rand()%4;
      int NS = 3+rnd.rand()%3;
      json vs;
      for (int j=0; j<NS; j++) {
        int M=2+rnd.rand()%2;
        set<int> S;
        for (int k=0; k<M; k++) S.insert(rnd.rand()%6-2);
        vs.push_back(S);
      }
      datain["vsetin"] = vs;
      datain["pred"] = (k==0? "sumzero" : k==1? "hassize2" : k==2? "hasodd" : "allpositive");
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> EJERC3 INSIDE
  typedef bool (*inside_t)(MatrixXd &xvrtx, VectorXd &xprobe);
  class eval3_t : public eval_base_t {
  public:
    inside_t F;
    eval3_t() { dumptests=0; F=NULL; ejerc=3; testfile="./in.json"; }
    void run_case(json &data,json &outuser) {
      vector<double> v = data["xvrtx"];
      vector<double> x = data["xprobe"];
      int ndim;
      if (v.size()==6) ndim=2;
      else if(v.size()==12) ndim=3;
      else throw string("bad size");
      MatrixXd xvrtx(ndim,ndim+1);
      VectorXd xprobe(ndim);
      int l=0;
      for (int j=0; j<ndim; j++) {
        xprobe(j) = x[j];
        for (int k=0; k<ndim+1; k++) 
          xvrtx(j,k) = v[l++];
      }
      bool z = F(xvrtx,xprobe);
      // cout << z << endl;
      outuser = z;
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int ndim = 2+rnd.rand()%2, nv=ndim*(ndim+1);
      int M = 10000;
      vector<double> xv(nv), xp(ndim);
#define XV(j,k) xv[(j)*(ndim+1)+k]
      for (int j=0; j<ndim; j++) {
        for (int k=0; k<ndim+1; k++) {
          XV(j,k) = rnd.rand()%M/double(M);
          xp[j] += XV(j,k);
        }
        xp[j] = xp[j]/(ndim+1) + 0.1*(rnd.rand()%M/double(M));
      }
      datain["xvrtx"] = xv;
      datain["xprobe"] = xp;
    }
  };
  
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  using Eval = eval_t<eval1_t,splitmono_t,
                      eval2_t,filtersets_t,
                      eval3_t,inside_t>;

}
#undef CSTR

#endif
