// GITVERSION: HPCMC.6-32-g014342ba
#ifndef EVAL_HPP
#define EVAL_HPP
#include <cassert>
#include <cstdlib>
#include <cstdlib>
#include <Eigen/Dense>
#include "aedtools/evalbase.hpp"

using namespace aed;
using namespace std;
using namespace Eigen;
#include "common.hpp"

bool sumzero(set<int> &);
bool hassize2(set<int> &);
bool hasodd(set<int> &);
bool allpositive(set<int> &);

namespace aed {

  void json2map(json &J,map<int,int> &M) {
    for (auto q:J) {
      int x = q[0],y = q[1];
      M[x] = y;
    }
  }

  void map2json(map<int,int> &M,json &J) {
    for (auto q:M) {
      vector<int> w;
      w.push_back(q.first);
      w.push_back(q.second);
      J.push_back(w);
    }
  }

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> EJERC1 MAPVEC
  static bool even(int x) { return x%2==0; }
  static bool odd(int x) { return x%2!=0; }
  static bool positive(int x) { return x>0; }
  static bool negative(int x) { return x<0; }
  static bool issqr(int x) {
    if (x<0) return false;
    int z=sqrt(x);
    return x==z*z;
  }
  typedef bool (*pred_t)(int);
  typedef void (*mapvec_t)(map<int,int> &M,pred_t pred,
                            vector<int> &v,vector<int> &w);
  class eval1_t : public eval_base_t {
  public:
    mapvec_t F;
    eval1_t() { dumptests=0; jsonvrbs=-1; F=NULL; ejerc=1; testfile="./ml.json"; }
    void run_case(json &data,json &outuser) {
      // Read v
      vector<int> v = data["v"], w;
      map<int,int> M;
      json2map(data["M"],M);
      string spred = data["pred"];
      pred_t pred=NULL;
#define Z(name) else if (spred==#name) pred=name
      if (0) {}
      Z(even);
      Z(odd);
      Z(positive);
      Z(negative);
      Z(issqr);
      else assert(0);
      F(M,pred,v,w);
      outuser = w;
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int N = 4+rnd.rand()%4;
      vector<int> v(N);
      for (int j=0; j<N; j++) v[j] = rnd.rand()%11-5;
      datain["v"] = v;
      int shift = rnd.rand()%5-2;
      map<int,int> M;
      for (int j=0; j<N; j++) { int x = v[j]; M[x] = x+shift; }
      json J;
      map2json(M,J);
      datain["M"] = J;
      int z = rnd.rand()%5;
      const char *table[] = {"even", "odd", "positive", "negative","issqr"};
      datain["pred"] = string(table[z]);
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> EJERC2 ONLY1
  typedef void (*only1_t)(vector< set<int> >&VS,set<int> &DS);
  class eval2_t : public eval_base_t {
  public:
    only1_t F;
    eval2_t() { dumptests=0; jsonvrbs=-1; F=NULL;
      ejerc=2; testfile="./ds.json"; }
    void run_case(json &data,json &outuser) {
      vector< set<int> > VS;
      json VSj = data;
      for (auto q:VSj) {
        set<int> S = q;
        VS.push_back(S);
      }
      set<int> DS;
      F(VS,DS);
      outuser = DS;
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int NS = 2+rnd.rand()%3;
      for (int j=0; j<NS; j++) {
        int m = 4+rnd.rand()%5;
        set<int> S;
        for (int k=0; k<m; k++) S.insert(rnd.rand()%15);
        datain.push_back(S);
      }
    }
  };
  
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> EJERC3 RECIPROCAL
  typedef bool (*reciprocal_t)(map<int,int> &M1,map<int,int> &M2);
  class eval3_t : public eval_base_t {
  public:
    reciprocal_t F;
    eval3_t() { dumptests=0; jsonvrbs=-1; F=NULL;
      ejerc=3; testfile="./rc.json"; }
    void run_case(json &data,json &outuser) {
      map<int,int> M1,M2;
      json2map(data["M1"],M1);
      json2map(data["M2"],M2);
      outuser["ret"] = F(M1,M2);
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int m = 4+rnd.rand()%5;
      set<int> S;
      for (int j=0; j<m; j++) S.insert(rnd.rand()%15);
      vector<int> x,y;
      int shift = rnd.rand()%4;
      for (auto q:S) {
        x.push_back(q);
        y.push_back(q+shift);
      }
      random_shuffle(rnd,x);
      random_shuffle(rnd,y);
      int n= x.size();
      map<int,int> M1,M2;
      // Probability of mutation
      int M = 2*n;
      for (int j=0; j<n; j++) {
        M1[x[j]] = y[j];
        int z = (rnd.rand()%M==0);
        M2[y[j]] = x[j]+z;
        
      }
      json M1j,M2j;
      map2json(M1,M1j);
      map2json(M2,M2j);
      datain["M1"] = M1j;
      datain["M2"] = M2j;
    }
  };

  void eigen2json(json &Aj,MatrixXd &A) {
    vector<int> sz(2);
    sz[0] = A.rows();
    sz[1] = A.cols();
    vector<double> v(sz[0]*sz[1]);
    int l=0;
    for (int j=0; j<sz[0]; j++)
      for (int k=0; k<sz[1]; k++)
        v[l++] = A(j,k);
    Aj.clear();
    Aj["size"] = sz;
    Aj["vals"] = v;
  }

  void json2eigen(MatrixXd &A,json &Aj) {
    vector<int> sz = Aj["size"];
    A.resize(sz[0],sz[1]);
    vector<double> v = Aj["vals"];
    int l=0;
    for (int j=0; j<sz[0]; j++)
      for (int k=0; k<sz[1]; k++)
        A(j,k) = v[l++];
  }

  void eigen2json(json &xj,VectorXd &x) {
    MatrixXd z = x;
    eigen2json(xj,z);
  }

  void json2eigen(VectorXd &x,json &xj) {
    MatrixXd z;
    json2eigen(z,xj);
    x = z;
  }

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> EJERC IS-INSIDECC
#undef EJERC
#undef EVAL
#define EJERC 4
#define EVAL eval4_t
#define SHV(x) cout << #x ": " << x << endl;
  typedef bool (*is_insidecc_t)(MatrixXd &X,VectorXd &xprobe);
  class EVAL : public eval_base_t {
  public:
    is_insidecc_t F;
    EVAL() { dumptests=0; jsonvrbs=-1; F=NULL;
      ejerc=EJERC; testfile="./iicc.json"; }
    void run_case(json &data,json &outuser) {
      MatrixXd X,z;
      VectorXd xprobe;
      json2eigen(X,data["X"]);
      json2eigen(z,data["xprobe"]);
      xprobe = z.col(0);
      // SHV(X);
      // SHV(xprobe);
      outuser["ret"] = F(X,xprobe);
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    double fix(double x) {
      const int M=100;
      return double(int(x*M))/M;
    }
    
    void generate_case(randomg_t &rnd,json &datain) {
      int ndim = 2+rnd.rand()%3;
      MatrixXd X(ndim,ndim+1), xprobe(ndim,1);
      VectorXd xc;
      for (int j=0; j<ndim; j++) {
        xprobe(j,0) = 0.4*(2.0*rnd.drand()-1.0);
        for (int k=0; k<ndim+1; k++) {
          X(j,k) = fix(rnd.drand());
        }
      }
      xc = X.rowwise().sum();
      xc /= (ndim+1);
      xprobe += xc;
      for (int j=0; j<ndim; j++) xprobe(j) = fix(xprobe(j));
      eigen2json(datain["X"],X);
      eigen2json(datain["xprobe"],xprobe);
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> EJERC TRUST-GAME
#undef EJERC
#undef EVAL
#define EJERC 5
#define EVAL eval5_t
#define SHV(x) cout << #x ": " << x << endl;
  typedef void (*trust_game_t)(map<string,void*> &data);
  class EVAL : public eval_base_t {
    class random_t : public player_t {
    public:
      randomg_t rnd;
      move_t move(vector<move_t> &me,vector<move_t> &you) {
        return (rnd.rand()%2==0? COLLAB : CHEAT);
      }
      string state() {return string("random"); }
    } random;

  public:
    trust_game_t F;
    EVAL() { dumptests=0; jsonvrbs=-1; F=NULL;
      ejerc=EJERC; testfile="./tg.json"; }
    void run_case(json &data,json &outuser) {
      map<string,void*> userdata;
      F(userdata);
      string base = data["type"];
      player_t *p = (player_t *)userdata[base];
      game_t G = (game_t)(userdata["game"]);
      int seed = data["seed"];
      random.rnd.reset(seed);
      if (p && G) {
        vector<move_t> mv1,mv2;
        int score1,score2;
        G(*p,random,20,mv1,mv2,score1,score2);
        outuser["you"] = mv1;
        outuser["other"] = mv2;
        outuser["yourscore"] = score1;
        outuser["otherscore"] = score2;
      }
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int z = rnd.rand()%5;
      if (0) {} // tricky
#define Q(q,type) else if (z==q) datain["type"] = #type
      Q(0,collabr);
      Q(1,cheater);
      Q(2,copycat);
      Q(3,copysmart);
      Q(4,grudger);
      datain["seed"] = rnd.rand();
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> EJERC TRUST-GAME
#undef EJERC
#undef EVAL
#define EJERC 6
#define EVAL eval6_t
  double poly(double x,int j) {
    if (j==0) return 1.0;
    else return pow(x,j);
  }

  double trigo(double x,int j) {
    int z=j%2, k=j/2;
    if (z==0) return cos(k*x);
    else return sin((k+1)*x);
  }

  double invpoly(double x,int j) {
    if (j==0) return 1.0;
    else return pow(x,-j);
  }

  typedef void (*lsqfit_t)(VectorXd &X,VectorXd &Y,basis_t basis,
                           int nterms,VectorXd &a);
  class EVAL : public eval_base_t {
  public:
    lsqfit_t F;
    EVAL() { dumptests=1; jsonvrbs=-1; F=NULL;
      ejerc=EJERC; testfile="./ls.json"; }
    void run_case(json &data,json &outuser) {
      MatrixXd z;
      VectorXd X,Y,a;
      basis_t basis=NULL;
      json2eigen(z,data["X"]); X = z;
      json2eigen(z,data["Y"]); Y = z;
      string sbasis = data["basis"];
      int nterms = data["nterms"];
      if (0) {}
#undef Q
#define Q(name) else if (sbasis==string(#name)) basis = name
      Q(poly);
      Q(trigo);
      Q(invpoly);
      F(X,Y,basis,nterms,a);
      eigen2json(outuser["a"],a);
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      MatrixXd auser,aref;
      json2eigen(aref,outref["a"]);
      json2eigen(auser,outuser["a"]);
      if (auser.rows()!=aref.rows()) return false;
      if (auser.cols()!=1) return false;
      VectorXd xuser=auser,xref=aref,da;
      da = xuser-xref;
      double ada = da.norm();
      printf("ada %g\n",ada);
      return ada<1e-4;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int z = rnd.rand()%3;
      basis_t basis;
      string sbasis;
#undef Q
#define Q(q,name) else if (z==q) { sbasis = string(#name); basis=name; }
      if (0) {}
      Q(0,poly)
      Q(1,trigo)
      Q(2,invpoly)
      int n=10;
      int nterms=4;
      VectorXd X(n),Y(n),a(nterms);
      for (int k=0; k<nterms; k++)
        a(k) = 2.0*rnd.drand()-1.0;

      for (int j=0; j<n; j++) {
        double x = 0.5+double(j)/n;
        X(j) = x;
        double Yj=0.0;
        for (int k=0; k<nterms; k++)
          Yj += a(k)*basis(x,k);
        Yj += 0.01*(2.0*rnd.drand()-1.0);
        Y(j) = Yj;
      }
      // SHV(X);
      // SHV(Y);
      // SHV(a);

      eigen2json(datain["X"],X);
      eigen2json(datain["Y"],Y);
      datain["nterms"] = nterms;
      datain["basis"] = sbasis;
    }
  };
  
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  using Eval = eval_t<eval1_t,mapvec_t,
                      eval2_t,only1_t,
                      eval3_t,reciprocal_t,
                      eval4_t,is_insidecc_t,
                      eval5_t,trust_game_t,
                      eval6_t,lsqfit_t>;

}
#undef CSTR

#endif
