// GITVERSION: CXXPROG.2-11-g5ff37152
#ifndef EVAL_HPP
#define EVAL_HPP
#include <cassert>
#include <cstdlib>
#include <cstdlib>
#include <Eigen/Dense>
#include "aedtools/evalbase.hpp"

using namespace aed;
using namespace std;
using namespace Eigen;

bool sumzero(set<int> &);
bool hassize2(set<int> &);
bool hasodd(set<int> &);
bool allpositive(set<int> &);

namespace aed {

  // ================================================================
  // STUFF FOR THIS EXAM
  // ================================================================
  void json2map(json &J,map<int,int> &M) {
    for (auto q:J) {
      int x = q[0],y = q[1];
      M[x] = y;
    }
  }

  void map2json(map<int,int> &M,json &J) {
    for (auto q:M) {
      vector<int> w;
      w.push_back(q.first);
      w.push_back(q.second);
      J.push_back(w);
    }
  }

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  typedef vector<int> (*findsum_t)(vector<int>& L, int S);
  class eval1_t : public eval_base_t {
  public:
    findsum_t F;
    eval1_t() { dumptests=0; F=NULL; ejerc=1; testfile="./ss.json"; }
    void run_case(json &data,json &outuser) {
      vector<int> L = data["L"];
      int S = data["S"];
      vector<int> R = F(L,S);
      outuser["R"] = R;
    }
    
    int check_case(json &datain,
                   json &outref,json &outuser) 
    {
      return outref==outuser;
    }
    
    void generate_case(randomg_t &rnd,json &datain) {
      vector<int> L;
      int n = 5+rnd.rand()%10, S=0;
      for(int i=0;i<n;i++) {
        int x = rnd.rand()%20;
        L.push_back(x);
        if (rnd.rand()%2==0) S+=x;
      }
      datain["L"] = L;
      datain["S"] = S;
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  typedef bool (*pred_int)(int);
  inline bool es_par(int x) { return x%2 == 0; }
  inline bool es_impar(int x) { return x%2 == 1; }
  // considerando que el generador genera ints entre 0 y 100
  inline bool es_menor_a_10(int x) { return x<10; }
  inline bool es_capicua(int x) { return x<10 || x/10==x%10; }
  inline bool es_primo(int x) {
    static set<int> primos =  {2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101};
    return primos.count(x)!=0;
  }
  inline bool es_compuesto(int x) { return !es_primo(x); }
  inline bool es_fibonacci(int x) {
    static set<int> primos =  {0,1,2,3,5,8,13,21,34,55,89,144};
    return primos.count(x)==0;
  }
  static map<string,pred_int > lista_predicados = { 
    {"es_par",es_par }, 
    {"es_impar",es_impar},
    {"es_primo",es_primo},
    {"es_compuesto",es_compuesto},
    {"es_capicua",es_capicua},
    {"es_menor_a_10",es_menor_a_10},
    {"es_fibonacci",es_fibonacci} };
  
  
  typedef int (*func_int)(int);
  inline int pow_2(int x) { return x*x; }
  inline int suma_5(int x) { return x+5; }
  inline int mod_5(int x) { return x%5; }
  inline int mod_10(int x) { return x%10; }
  inline int div_int_5(int x) { return x/5; }
  inline int div_int_10(int x) { return x/10; }
  inline int ordena_digitos(int x) { 
    std::stringstream ss; ss<<x; 
    std::string s = ss.str();
    std::sort(s.begin(),s.end());
    return std::atoi(s.c_str());
  }
  static map<string,func_int> lista_funciones= { 
    {"pow_2",pow_2 }, 
    {"suma_5",suma_5},
    {"mod_5",mod_5 },
    {"mod_10",mod_10},
    {"div_int_5",div_int_5},
    {"div_int_10",div_int_10},
    {"ordena_digitos",ordena_digitos}
  };
  
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  vector<set<int> > readLS(json &data, std::string name) { 
    vector<set<int> > LS;
    json jLS = data[name];
    for(auto p = jLS.begin(); p!=jLS.end(); p++)
      LS.push_back(*p);
    return LS;
  }
  
  void writeLS(json &data, std::string name, vector<set<int> > &LS) { 
    json &jLS = data[name];
    for(auto &S : LS)
      jLS.push_back(S);
  }
  
  typedef void (*xinyect_t)(vector<set<int>> &LS_IN,
                                int (*func)(int),vector<set<int> > &LS_OUT);
  class eval_xinyect_t : public eval_base_t {
  public:
    xinyect_t F;
    eval_xinyect_t() { dumptests=0; F=NULL; ejerc=2; testfile="./fs.json"; }
    void run_case(json &data,json &outuser) {
      string fname = data["fname"];
      vector<set<int> > LS_IN = readLS(data,"LS_IN"), LS_OUT;
      F(LS_IN,lista_funciones[fname],LS_OUT);
      writeLS(outuser,"LS_OUT",LS_OUT);
    }
    
    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }
    
    void generate_case(randomg_t &rnd,json &datain) {
      vector<set<int> > LS(3+rnd.rand()%5);
      for (auto &S: LS) {
        // int Nelems = 3+rnd.rand()%8;
        for (int k=0; k<10; k++) S.insert(rnd.rand()%25);
      }
      writeLS(datain,"LS_IN",LS);
      auto it = lista_funciones.begin();
      advance(it,rnd.rand()%lista_funciones.size());
      datain["fname"] = it->first;
    }
  };

  
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  using Eval = eval_t<eval1_t,findsum_t,
                      eval_xinyect_t,xinyect_t>;

}
#undef CSTR

#endif
